/*
 * Copyright (C) 2011 Eiichiro Uchiumi. All Rights Reserved.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 *     
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.eiichiro.acidhouse.appengine;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.logging.Logger;

import org.eiichiro.acidhouse.IndoubtException;
import org.eiichiro.acidhouse.Log;
import org.eiichiro.acidhouse.Log.Operation;
import org.eiichiro.acidhouse.Log.State;
import org.eiichiro.reverb.lang.UncheckedException;

import com.google.appengine.api.datastore.Transaction;

/**
 * {@code AppEngineTransaction} is a Google App Engine Low-level Datastore API 
 * based implementation of {@code Transaction}.
 * 
 * @author <a href="mailto:eiichiro@eiichiro.org">Eiichiro Uchiumi</a>
 */
public class AppEngineTransaction implements org.eiichiro.acidhouse.Transaction {

	private final Logger logger = Logger.getLogger(getClass().getName());
	
	private final String id;
	
	private AppEngineDatastoreSession session;
	
	protected List<Log> logs = new ArrayList<Log>();
	
	protected Map<Object, Transaction> transactions = new HashMap<Object, Transaction>();
	
	/**
	 * Constructs a new {@code AppEngineTransaction} instance with random UUID 
	 * based transaction id and the specified {@code AppEngineDatastoreSession} 
	 * instance.
	 * 
	 * @param session {@code AppEngineDatastoreSession} instance.
	 */
	public AppEngineTransaction(AppEngineDatastoreSession session) {
		this(UUID.randomUUID().toString(), session);
	}
	
	/**
	 * Constructs a new {@code AppEngineTransaction} instance with the specified 
	 * transaction id and the specified {@code AppEngineDatastoreSession} 
	 * instance.
	 * 
	 * @param id The transaction id.
	 * @param session {@code AppEngineDatastoreSession} instance.
	 */
	public AppEngineTransaction(String id, AppEngineDatastoreSession session) {
		this.id = id;
		this.session = session;
	}

	/* (non-Javadoc)
	 * @see org.eiichiro.acidhouse.appengine.DatastoreTransaction#getId()
	 */
	@Override
	public String getId() {
		return id;
	}

	/** Commits this transaction with loose commitment protocol. */
	@Override
	public void commit() throws IndoubtException {
		session.transaction.remove();
		
		if (logs.size() == 0) {
			return;
		}
		
		int committed = 0;
		
		try {
			for (Log log : logs) {
				if (log.getOperation() != Operation.GET) {
					Transaction tx = transactions.get(log.getEntity());
					tx.commit();
					log.setState(State.COMMITTED);
					committed++;
				}
			}

		} catch (Exception e) {
			if (committed > 0) {
				throw new IndoubtException(e, id, logs);
			} else {
				throw new UncheckedException(e);
			}
		}
		
		logger.info("Transaction [" + id + "] committed");
	}

	/** Rolls back this transaction. */
	@Override
	public void rollback() {
		session.transaction.remove();
		
		for (Transaction transaction : transactions.values()) {
			if (transaction.isActive()) {
				transaction.rollback();
			}
		}
	}
	
}
